use std::{
    io::Write,
    sync::{Arc, RwLock},
};

use serde::{Deserialize, Serialize};
use thiserror::Error;

use crate::{
    dirs::{Dirs, DirsError},
    models::{CodeChallengeMethod, CodeVerifier},
};

#[derive(Default, Debug, Clone, Serialize, Deserialize)]
pub struct Config {
    #[serde(with = "arc_rwlock_serde")]
    token: Arc<RwLock<Option<String>>>,
    #[serde(skip)]
    code_verifier: Arc<RwLock<Option<CodeVerifier>>>,
    #[serde(skip)]
    code_challenge_method: Arc<RwLock<Option<CodeChallengeMethod>>>,
}

impl PartialEq for Config {
    fn eq(&self, _: &Self) -> bool {
        true
    }
}

#[derive(Debug, Error)]
pub enum ConfigError {
    #[error(transparent)]
    Io(#[from] std::io::Error),
    #[error(transparent)]
    Dirs(#[from] DirsError),
    #[error(transparent)]
    Toml(#[from] toml::ser::Error),
    #[error(transparent)]
    Unknown(#[from] anyhow::Error),
}

impl Config {
    pub fn set_token(&self, token: Option<String>) -> Result<(), ConfigError> {
        *self.token.write().unwrap() = token;
        self.write()?;
        Ok(())
    }

    pub fn token(&self) -> Option<String> {
        self.token.read().unwrap().clone()
    }
}

impl Config {
    pub fn set_code_verifier(
        &self,
        code_verifier: Option<CodeVerifier>,
    ) -> Result<(), ConfigError> {
        *self.code_verifier.write().unwrap() = code_verifier;
        Ok(())
    }

    pub fn set_code_challenge_method(
        &self,
        code_challenge_method: Option<CodeChallengeMethod>,
    ) -> Result<(), ConfigError> {
        *self.code_challenge_method.write().unwrap() = code_challenge_method;
        Ok(())
    }

    pub fn code_verifier(&self) -> Option<CodeVerifier> {
        self.code_verifier.read().unwrap().clone()
    }

    pub fn code_challenge_method(&self) -> Option<CodeChallengeMethod> {
        self.code_challenge_method.read().unwrap().clone()
    }
}

impl Config {
    pub(crate) fn new() -> Result<Self, ConfigError> {
        let config = config::Config::builder()
            .add_source(
                config::File::with_name(Dirs::get_config_file()?.to_str().unwrap()).required(false),
            )
            .build()
            .unwrap();

        let config: Self = config.try_deserialize().unwrap_or_default();

        config.write()?;

        Ok(config)
    }

    fn write(&self) -> Result<(), ConfigError> {
        let toml = toml::to_string_pretty(self)?;
        let mut file = std::fs::File::create(&Dirs::get_config_file()?)?;
        file.write_all(toml.as_bytes())?;
        Ok(())
    }
}

mod arc_rwlock_serde {
    use serde::de::Deserializer;
    use serde::ser::Serializer;
    use serde::{Deserialize, Serialize};
    use std::sync::{Arc, RwLock};

    pub fn serialize<S, T>(val: &Arc<RwLock<T>>, s: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
        T: Serialize,
    {
        T::serialize(&*val.read().unwrap(), s)
    }

    pub fn deserialize<'de, D, T>(d: D) -> Result<Arc<RwLock<T>>, D::Error>
    where
        D: Deserializer<'de>,
        T: Deserialize<'de>,
    {
        Ok(Arc::new(RwLock::new(T::deserialize(d)?)))
    }
}
